#include "febmap.h"
#include "trace.h"
#include <cmath>
#include <algorithm>
#include <numeric>

using namespace std;

FEBMAP::FEBMAP(std::vector<int> brOrders, int typeNum) : MixtureDensityMAP(typeNum), M(brOrders.size()) {

    branchOrders = new int[M];
    copy (brOrders.begin(), brOrders.end(), branchOrders);
    lambda = new double[M];
    alpha  = new double[M];
    z  = new double[M];
    P = new double[C*M*M];
}

FEBMAP::~FEBMAP() {

    delete [] branchOrders;
    delete [] lambda;
    delete [] alpha;
    delete [] z;
    delete [] P;
}

void FEBMAP::setInitialGuess(InitialGuess initial, Trace *trace, double trMean) {

    if (initial == IGKMeans) {
        vector<char> clusters = trace->kmeans (M);
        vector<double> means (M);
        vector<double> vars (M);
        vector<int> nums (M);
        vector<int> index (M);
        fill (means.begin(), means.end(), 0.0);
        fill (vars.begin(), vars.end(), 0.0);
        fill (nums.begin(), nums.end(),0);
        for (int k=0; k<clusters.size(); k++) {
            means[clusters[k]] += trace->getIATime(k);
            vars[clusters[k]] += trace->getIATime(k) * trace->getIATime(k);
            nums[clusters[k]]++;
        }
        for (int i=0; i<M; i++) {
            means[i] /= nums[i];
            vars[i] = vars[i]/nums[i]/means[i]/means[i] - 1.0;
        }
        // Sort means with respect to the variances. Simple dumb bubble sort.
        for (int i=0; i<M; i++)
            index[i] = i;
        for (int a=0; a<M; a++)
            for (int b=0; b<M-1-a; b++)
                if (vars[b+1]<vars[b]) {
                    swap (vars[b], vars[b+1]);
                    swap (means[b], means[b+1]);
                    swap (index[b], index[b+1]);
                }
        // assign rates and initial probabilties to the branches
        for (int i=0; i<M; i++) {
            alpha[i] = (double) nums[i] / clusters.size();
            lambda[i] = 1.0/means[i] * branchOrders[i];
            z[i] = (double) rand() / RAND_MAX;
        }
        setSource("kmeans");
    }
    else if (initial == IGLinear) {
        double herMean = 0.0;
        for (int i=0; i<M; i++) {
            alpha[i] = 1.0 / M;
            lambda[i] = (i+1.0) * branchOrders[i];
            z[i] = 0.5;
            herMean += alpha[i] * 2.0 * branchOrders[i] / lambda[i];
        }
        for (int i=0; i<M; i++)
            lambda[i] *= herMean/trMean;
        setSource("linear");
    }
    else if (initial == IGRandom) {
        double alsum = 0.0;
        for (int i=0; i<M; i++) {
            alsum += alpha[i] = rand();
            lambda[i] = branchOrders[i] / trace->getIATime(rand() % trace->size());
            z[i] = (double) rand() / RAND_MAX;
        }
        for (int i=0; i<M; i++)
            alpha[i] /= alsum;
        setSource("random");
    }
    else if (initial == IGExponential) {
        double herMean = 0.0;
        for (int i=0; i<M; i++) {
            alpha[i] = 1.0 / M;
            lambda[i] = pow(10.0, i) * branchOrders[i];
            z[i] = 0.5;
            herMean += alpha[i] * 2.0 * branchOrders[i] / lambda[i];
        }
        for (int i=0; i<M; i++)
            lambda[i] *= herMean/trMean;
        setSource("exponential");
    }
    // initial value for P
    for (int c=0; c<C; c++)
        for (int i=0; i<M; i++)
            for (int j=0; j<M; j++)
                P[c*M*M+i*M+j] = alpha[j]/C;
}

std::vector<double> FEBMAP::getBranchMoments(int order) const {

    vector<double> moms (M, 0.0);
    for (int i=0; i<M; i++) {
        double brp = 1.0 - z[i];
        int k = 1;
        while (brp > 1e-14) {
            // moment of an Erlang-k*branchOrders[i] distribution
            double mom = 1.0 / pow (lambda[i], order);
            for (int j=0; j<order; j++)
                mom *= k*branchOrders[i]+j;
            moms[i] += brp * mom;
            brp *= z[i];
            k++;
        }
    }
    return moms;
}

void FEBMAP::printSummary(ostream &os) const {

    os << "LogLi = " << logli << " (";
    for (int i=0; i<M; i++)
        os << branchOrders[i] << ", ";
    os << source << ")" << endl;
}

void FEBMAP::printAsPH (std::ostream& os, const std::string& format) const {

    vector<vector<double> > D0;
    int siz = accumulate (branchOrders, branchOrders+M, 0);
    D0.resize(siz);
    for (int i=0; i<siz; i++)
        D0[i].resize(siz);

    vector<double> alp (siz, 0.0);
    // fill with zero
    for_each(D0.begin(), D0.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0, blstart = 0;
    for (int i=0; i<M; i++) {
        alp[blstart] = alpha[i];
        for (int b=0; b<branchOrders[i]; b++) {
            D0[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                D0[ix][ix+1] = lambda[i];
            else
                D0[ix][blstart] += z[i]*lambda[i];
            ix++;
        }
        blstart = ix;
    }

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        os << "A = ";
        printMatrix (os, D0, format);
        os << endl << "alpha = ";
        printVector (os, alp, format);
        os << endl;
    }
    else if (format=="numpy") {
        os << "A = matrix(";
        printMatrix(os, D0, format);
        os << ")" << endl << "alpha = matrix([";
        printVector (os, alp, format);
        os << "])" << endl;
    }
}

void FEBMAP::printAsMAP (ostream &os, const string &format) const {

    vector<vector<vector<double> > > D;

    vector<vector<double> > mat;
    int siz = accumulate (branchOrders, branchOrders+M, 0);
    mat.resize(siz);
    for (int i=0; i<siz; i++)
        mat[i].resize(siz);

    // Matrix D0
    // ---------
    // fill with zero
    for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0, blstart = 0;
    for (int i=0; i<M; i++) {
        for (int b=0; b<branchOrders[i]; b++) {
            mat[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                mat[ix][ix+1] = lambda[i];
            else
                mat[ix][blstart] += z[i]*lambda[i];
            ix++;
        }
        blstart = ix;
    }
    D.push_back(mat);

    // Matrix Dk
    // ---------
    vector<int> startState, endState;
    for (int i=0; i<M; i++) {
        startState.push_back(i==0 ? 0 : startState[i-1] + branchOrders[i-1]);
        endState.push_back(startState[i] + branchOrders[i] - 1);
    }

    for (int c=0; c<C; c++) {
        // fill with zero
        for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
        ix = 0;
        for (int i=0; i<M; i++)
            for (int j=0; j<M; j++)
                mat[endState[i]][startState[j]] = (1.0-z[i]) * lambda[i] * P[c*M*M+i*M+j];
        D.push_back(mat);
    }

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = ";
            printMatrix(os, D[k], format);
            os << endl;
        }
    }
    else if (format=="numpy") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = matrix(";
            printMatrix(os, D[k], format);
            os << ")" << endl;
        }
    }
}
