//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "hermap.h"
#include "trace.h"
#include <limits>
#include <cmath>
#include <cstring>
#include <algorithm>
#include <numeric>

using namespace std;

HErMAP::HErMAP(std::vector<int> brOrders, int typeNum) : MixtureDensityMAP(typeNum), M(brOrders.size()) {

    branchOrders = new int[M];
    copy (brOrders.begin(), brOrders.end(), branchOrders);
    lambda = new double[M];
    alpha  = new double[M];
    P = new double[C*M*M];
}

HErMAP::~HErMAP() {

    delete [] branchOrders;
    delete [] lambda;
    delete [] alpha;
    delete [] P;
}

void HErMAP::setInitialGuess(InitialGuess initial, Trace *trace, double trMean) {

    if (initial == IGKMeans) {
        vector<char> clusters = trace->kmeans (M);
        vector<double> means (M);
        vector<double> vars (M);
        vector<int> nums (M);
        vector<int> index (M);
        fill (means.begin(), means.end(), 0.0);
        fill (vars.begin(), vars.end(), 0.0);
        fill (nums.begin(), nums.end(),0);
        for (int k=0; k<clusters.size(); k++) {
            means[clusters[k]] += trace->getIATime(k);
            vars[clusters[k]] += trace->getIATime(k) * trace->getIATime(k);
            nums[clusters[k]]++;
        }
        for (int i=0; i<M; i++) {
            means[i] /= nums[i];
            vars[i] = vars[i]/nums[i]/means[i]/means[i] - 1.0;
        }
        // Sort means with respect to the variances. Simple dumb bubble sort.
        for (int i=0; i<M; i++)
            index[i] = i;     
        for (int a=0; a<M; a++)
            for (int b=0; b<M-1-a; b++)
                if (vars[b+1]<vars[b]) {
                    swap (vars[b], vars[b+1]);
                    swap (means[b], means[b+1]);
                    swap (index[b], index[b+1]);
                }
        // assign rates and initial probabilties to the branches
        for (int i=0; i<M; i++) {
            alpha[i] = (double) nums[i] / clusters.size();
            lambda[i] = 1.0/means[i] * branchOrders[i];
        }       
        setSource("kmeans");
    }
    else if (initial == IGLinear) {
        double herMean = 0.0;
        for (int i=0; i<M; i++) {
            alpha[i] = 1.0 / M;
            lambda[i] = (i+1.0) * branchOrders[i];
            herMean += alpha[i] * branchOrders[i] / lambda[i];
        }
        for (int i=0; i<M; i++)
            lambda[i] *= herMean/trMean;
        setSource("linear");
    }
    else if (initial == IGRandom) {
        double herMean = 0.0;
        double alsum = 0.0;
        for (int i=0; i<M; i++) {
            alsum += alpha[i] = rand();
            lambda[i] = branchOrders[i] / trace->getIATime(rand() % trace->size());
        }
        for (int i=0; i<M; i++)
            alpha[i] /= alsum;
        setSource("random");
    }
    else if (initial == IGExponential) {
        double herMean = 0.0;
        for (int i=0; i<M; i++) {
            alpha[i] = 1.0 / M;
            lambda[i] = pow(10.0, i) * branchOrders[i];
            herMean += alpha[i] * branchOrders[i] / lambda[i];
        }
        for (int i=0; i<M; i++)
            lambda[i] *= herMean/trMean;
        setSource("exponential");
    }
    // initial value for P
    for (int c=0; c<C; c++)
        for (int i=0; i<M; i++)
            for (int j=0; j<M; j++)
                P[c*M*M+i*M+j] = alpha[j]/C;
}

std::vector<double> HErMAP::getBranchMoments(int order) const {

    vector<double> moms (M);
    for (int i=0; i<M; i++) {
        moms[i] = 1.0 / pow (lambda[i], order);
        for (int j=0; j<order; j++)
            moms[i] *= branchOrders[i]+j;
    }
    return moms;
}

void HErMAP::printSummary(ostream &os) const {

    os << "LogLi = " << logli << " (";
    for (int i=0; i<M; i++)
        os << branchOrders[i] << ", ";
    os << source << ")" << endl;
}

void HErMAP::printAsPH (std::ostream& os, const std::string& format) const {

    vector<vector<double> > D0;
    int siz = accumulate (branchOrders, branchOrders+M, 0);
    D0.resize(siz);
    for (int i=0; i<siz; i++)
        D0[i].resize(siz);

    // Matrix D0
    // ---------
    // fill with zero
    for_each(D0.begin(), D0.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0;
    for (int i=0; i<M; i++)
        for (int b=0; b<branchOrders[i]; b++) {
            D0[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                D0[ix][ix+1] = lambda[i];
            ix++;
        }

    vector<double> alp (alpha, alpha+M);

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        os << "A = ";
        printMatrix (os, D0, format);
        os << endl << "alpha = ";
        printVector (os, alp, format);
        os << endl;
    }
    else if (format=="numpy") {
        os << "A = matrix(";
        printMatrix(os, D0, format);
        os << ")" << endl << "alpha = matrix([";
        printVector (os, alp, format);
        os << "])" << endl;
    }
}

void HErMAP::printAsMAP (ostream &os, const string &format) const {

    vector<vector<vector<double> > > D;

    vector<vector<double> > mat;
    int siz = accumulate (branchOrders, branchOrders+M, 0);
    mat.resize(siz);
    for (int i=0; i<siz; i++)
        mat[i].resize(siz);

    // Matrix D0
    // ---------
    // fill with zero
    for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0;
    for (int i=0; i<M; i++)
        for (int b=0; b<branchOrders[i]; b++) {
            mat[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                mat[ix][ix+1] = lambda[i];
            ix++;
        }
    D.push_back(mat);

    // Matrix Dk
    // ---------
    vector<int> startState, endState;
    for (int i=0; i<M; i++) {
        startState.push_back(i==0 ? 0 : startState[i-1] + branchOrders[i-1]);
        endState.push_back(startState[i] + branchOrders[i] - 1);
    }

    for (int c=0; c<C; c++) {
        // fill with zero
        for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
        ix = 0;
        for (int i=0; i<M; i++)
            for (int j=0; j<M; j++)
                mat[endState[i]][startState[j]] = lambda[i] * P[c*M*M+i*M+j];
        D.push_back(mat);
    }

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = ";
            printMatrix(os, D[k], format);
            os << endl;
        }
    }
    else if (format=="numpy") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = matrix(";
            printMatrix(os, D[k], format);
            os << ")" << endl;
        }
    }
}
