//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "mixturedensmap.h"
#include <limits>
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iostream>

using namespace std;

MixtureDensityMAP::MixtureDensityMAP (int typeNum)
    : C(typeNum), source("unknown"), logli(-std::numeric_limits<double>::infinity()), iter(0) {
}

vector<double> MixtureDensityMAP::getMoments(int upToOrder) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();

    vector<double> moms(upToOrder+1, 0.0);
    for (int i=0; i<=upToOrder; i++) {
        vector<double> bmoms = getBranchMoments(i);
        for (int j=0; j<N; j++)
            moms[i] += alpha[j] * bmoms[j];
    }
    return moms;
}

std::vector<double> MixtureDensityMAP::getBackwardMoments(int upToOrder, int k) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    vector<double> moms(upToOrder+1, 0.0);

    double* p = new double[N];
    for (int j=0; j<N; j++) {
        p[j] = 0.0;
        for (int i=0; i<N; i++)
            p[j] += P[k*N*N+j*N+i];
    }

    for (int i=0; i<=upToOrder; i++) {
        vector<double> bmoms = getBranchMoments(i);
        double norm = 0.0;
        for (int j=0; j<N; j++) {
            moms[i] += alpha[j] * bmoms[j] * p[j];
            norm += alpha[j] * p[j];
        }
        moms[i] /= norm;
    }

    delete [] p;
    return moms;
}

std::vector<double> MixtureDensityMAP::getForwardMoments(int upToOrder, int k) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    vector<double> moms(upToOrder+1, 0.0);

    double* p = new double[N];
    for (int i=0; i<N; i++) {
        p[i] = 0.0;
        for (int j=0; j<N; j++)
            p[i] += alpha[j] * P[k*N*N+j*N+i];
    }

    for (int i=0; i<=upToOrder; i++) {
        vector<double> bmoms = getBranchMoments(i);
        double norm = 0.0;
        for (int j=0; j<N; j++) {
            moms[i] += p[j] * bmoms[j];
            norm += p[j];
        }
        moms[i] /= norm;
    }

    delete [] p;

    return moms;
}

vector<vector<double> > MixtureDensityMAP::getJointMoments(int upToOrder) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    double* sumP = new double[N*N];
    memset (sumP, 0, N*N*sizeof(double));
    for (int i=0; i<N*N; i++)
        for (int c=0; c<C; c++)
            sumP[i] += P[c*N*N+i];

    vector<vector<double> > jmoms (upToOrder+1);
    for (int i=0; i<=upToOrder; i++) {
        jmoms[i].resize(upToOrder+1);
        fill (jmoms[i].begin(), jmoms[i].end(), 0.0);
    }

    vector<vector<double> > bmoms;
    for (int i=0; i<=upToOrder; i++)
        bmoms.push_back (getBranchMoments(i));

    for (int n1=0; n1<=upToOrder; n1++) {
        for (int n2=0; n2<=upToOrder; n2++) {
            for (int i=0; i<N; i++)
                for (int j=0; j<N; j++)
                    jmoms[n1][n2] += alpha[i]*sumP[i*N+j]*bmoms[n1][i]*bmoms[n2][j];
        }
    }
    delete [] sumP;
    return jmoms;
}

std::vector<std::vector<double> > MixtureDensityMAP::getJointMomentsOfType(int upToOrder, int p) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    vector<vector<double> > jmoms (upToOrder+1);
    for (int i=0; i<=upToOrder; i++) {
        jmoms[i].resize(upToOrder+1);
        fill (jmoms[i].begin(), jmoms[i].end(), 0.0);
    }

    vector<vector<double> > bmoms;
    for (int i=0; i<=upToOrder; i++)
        bmoms.push_back (getBranchMoments(i));

    for (int n1=0; n1<=upToOrder; n1++) {
        for (int n2=0; n2<=upToOrder; n2++) {
            for (int i=0; i<N; i++)
                for (int j=0; j<N; j++)
                    jmoms[n1][n2] += alpha[i]*P[p*N*N+i*N+j]*bmoms[n1][i]*bmoms[n2][j];
        }
    }
    return jmoms;
}

vector<double> MixtureDensityMAP::getAcf (int maxLag) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    double* sumP = new double[N*N];
    memset (sumP, 0, N*N*sizeof(double));
    for (int i=0; i<N*N; i++)
        for (int c=0; c<C; c++)
            sumP[i] += P[c*N*N+i];

    vector<double> acf;
    vector<double> bmeans = getBranchMoments (1);
    vector<double> moms = getMoments(2);
    double* Ppow = new double[N*N];
    double* tmp = new double[N*N];

    // set ppow to identity matrix
    memset (Ppow, 0.0, N*N*sizeof(double));
    for (int i=0; i<N; i++)
        Ppow[i*N+i] = 1.0;

    // calculate lag-k acf
    for (int k=0; k<maxLag; k++) {
        // calculate lag-k joint mean
        double jm = 0.0;
        for (int i=0; i<N; i++)
            for (int j=0; j<N; j++)
                jm += alpha[i]*Ppow[i*N+j]*bmeans[i]*bmeans[j];
        // acf
        acf.push_back ((jm-moms[1] * moms[1]) / (moms[2] - moms[1]*moms[1]));
        // tmp = ppow * p
        memset (tmp, 0.0, N*N*sizeof(double));
        for (int i=0; i<N; i++)
            for (int j=0; j<N; j++)
                for (int x=0; x<N; x++)
                    tmp[i*N+j] += Ppow[i*N+x] * sumP[x*N+j];
        // ppow = tmp
        memcpy (Ppow, tmp, N*N*sizeof(double));
    }

    delete [] Ppow;
    delete [] tmp;
    delete [] sumP;

    return acf;
}

std::vector<double> MixtureDensityMAP::getCrossAcf(int maxLag, int p1, int p2) const {

    int N = getNumOfBranches();
    const double* alpha = getBranchProbs ();
    const double* P = getBranchTransitionMatrix();

    double* sumP = new double[N*N];
    memset (sumP, 0, N*N*sizeof(double));
    for (int i=0; i<N*N; i++)
        for (int c=0; c<C; c++)
            sumP[i] += P[c*N*N+i];

    vector<double> acf;
    vector<double> bmeans = getBranchMoments (1);
    vector<double> moms = getMoments (2);
    vector<double> moms1 = getBackwardMoments (2, p1);
    vector<double> moms2 = getBackwardMoments (2, p2);
    double std1 = sqrt (moms1[2] - moms1[1]*moms1[1]);
    double std2 = sqrt (moms2[2] - moms2[1]*moms2[1]);

    double* Ppow = new double[N*N];
    double* tmp = new double[N*N];

    // set ppow to identity matrix
    memset (Ppow, 0.0, N*N*sizeof(double));
    for (int i=0; i<N; i++)
        Ppow[i*N+i] = 1.0;

    // opening vectors
    double* a0 = new double[N];
    double* a1 = new double[N];
    for (int i=0; i<N; i++) {
        a0[i] = 0.0; a1[i] = 0.0;
        for (int j=0; j<N; j++) {
            a0[i] += alpha[j]*P[p1*N*N+j*N+i];
            a1[i] += alpha[j]*bmeans[j]*P[p1*N*N+j*N+i];
        }
    }
    // closing vectors
    double* b0 = new double[N];
    double* b1 = new double[N];
    for (int i=0; i<N; i++) {
        b0[i] = 0.0; b1[i] = 0.0;
        for (int j=0; j<N; j++) {
            b0[i] += P[p2*N*N+i*N+j];
            b1[i] += bmeans[i]*P[p2*N*N+i*N+j];
        }
    }

    // calculate lag-k acf
    for (int k=0; k<maxLag; k++) {
        // calculate lag-k joint mean
        double jm1 = 0.0, jm2 = 0.0, jm3 = 0.0, jm4 = 0.0;
        for (int i=0; i<N; i++)
            for (int j=0; j<N; j++) {
                jm1 += a1[i]*Ppow[i*N+j]*b1[j];
                jm2 += a1[i]*Ppow[i*N+j]*b0[j];
                jm3 += a0[i]*Ppow[i*N+j]*b1[j];
                jm4 += a0[i]*Ppow[i*N+j]*b0[j];
            }
        // acf
        acf.push_back( (jm1 - jm2*moms2[1] - moms1[1]*jm3 + moms1[1]*moms2[1]*jm4) / jm4 / std1 / std2 );
        // tmp = ppow * p
        memset (tmp, 0.0, N*N*sizeof(double));
        for (int i=0; i<N; i++)
            for (int j=0; j<N; j++)
                for (int x=0; x<N; x++)
                    tmp[i*N+j] += Ppow[i*N+x] * sumP[x*N+j];
        // ppow = tmp
        memcpy (Ppow, tmp, N*N*sizeof(double));
    }
    delete [] Ppow;
    delete [] tmp;
    delete [] sumP;
    delete [] a0;
    delete [] a1;
    delete [] b0;
    delete [] b1;

    return acf;
}

void MixtureDensityMAP::printVector(ostream &os, std::vector<double> vec, string format) const {

    if (format=="text")
        for (int j=0; j<vec.size(); j++)
            os << vec[j] << '\t';
    else if (format=="matlab") {
        os << "[";
        for (int j=0; j<vec.size(); j++) {
            os << vec[j];
            if (j<vec.size()-1)
                os << ", ";
        }
        os << "]";
    }
    else if (format=="mathematica") {
        os << "{";
        for (int j=0; vec.size(); j++) {
            os << vec[j];
            if (j<vec.size()-1)
                os << ", ";
        }
        os << "}";
    }
    else if (format=="numpy") {
        os << "[";
        for (int j=0; j<vec.size(); j++) {
            os << vec[j];
            if (j<vec.size()-1)
                os << ", ";
        }
        os << "]";
    }
}

void MixtureDensityMAP::printMatrix (std::ostream& os, std::vector<std::vector<double> > mat, std::string format) const {

    if (format=="text") {
        for (int i=0; i<mat.size(); i++) {
            printVector(os, mat[i], format);
            os << endl;
        }
    }
    else if (format=="matlab") {
        os << "[";
        for (int i=0; i<mat.size(); i++) {
            for (int j=0; j<mat[i].size(); j++) {
                os << mat[i][j];
                if (j<mat[i].size()-1)
                    os << ", ";
            }
            if (i<mat.size()-1)
                os << "; ";
        }
        os << "]";
    }
    else if (format=="mathematica") {
        os << "{";
        for (int i=0; i<mat.size(); i++) {
            printVector(os, mat[i], format);
            if (i<mat.size()-1)
                os << ", ";
        }
        os << "}";
    }
    else if (format=="numpy") {
        os << "[";
        for (int i=0; i<mat.size(); i++) {
            printVector(os, mat[i], format);
            if (i<mat.size()-1)
                os << ", ";
        }
        os << "])" << endl;
    }
}
