//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "specmap.h"
#include "trace.h"
#include <limits>
#include <cmath>
#include <cstring>
#include <algorithm>
#include <numeric>

using namespace std;

SpecMAP::SpecMAP(std::vector<int> brOrders, int typeNum)
    : MixtureDensityMAP(typeNum), M(brOrders.size()) {

    branchOrders = new int[M];
    copy (brOrders.begin(), brOrders.end(), branchOrders);
    N = accumulate (brOrders.begin(), brOrders.end(), 0);
    lambda = new double[M];
    alpha = new double[N];
    P = new double[C*M*N];
    bigP = NULL;
}

SpecMAP::~SpecMAP() {

    delete [] branchOrders;
    delete [] lambda;
    delete [] alpha;
    delete [] P;
    delete [] bigP;
}

void SpecMAP::setInitialGuess(InitialGuess initial, Trace *trace, double trMean) {

    if (initial == IGKMeans) {
        vector<char> clusters = trace->kmeans (M);
        vector<double> means (M);
        vector<double> vars (M);
        vector<int> nums (M);
        vector<int> index (M);
        fill (means.begin(), means.end(), 0.0);
        fill (vars.begin(), vars.end(), 0.0);
        fill (nums.begin(), nums.end(),0);
        for (int k=0; k<clusters.size(); k++) {
            means[clusters[k]] += trace->getIATime(k);
            vars[clusters[k]] += trace->getIATime(k) * trace->getIATime(k);
            nums[clusters[k]]++;
        }
        for (int i=0; i<M; i++) {
            means[i] /= nums[i];
            vars[i] = vars[i]/nums[i]/means[i]/means[i] - 1.0;
        }

        // Sort means with respect to the variances. Simple dumb bubble sort.
        for (int i=0; i<M; i++)
            index[i] = i;
        for (int a=0; a<M; a++)
            for (int b=0; b<M-1-a; b++)
                if (vars[b+1]<vars[b]) {
                    swap (vars[b], vars[b+1]);
                    swap (means[b], means[b+1]);
                    swap (index[b], index[b+1]);
                }
        // assign rates and initial probabilties to the branches
        int ix = 0;
        for (int i=0; i<M; i++) {
            for (int j=0; j<branchOrders[i]; j++)
                alpha[ix++] = (double) nums[i] / clusters.size() / branchOrders[i];
            lambda[i] = 0.0;
            for (int j=0; j<branchOrders[i]; j++)
                lambda[i] = 1.0/means[i] * 1.0 / (j+1.0);
        }
        setSource("kmeans");
    }
    else if (initial == IGLinear) {
        for (int i=0; i<N; i++)
            alpha[i] = 1.0 / N;
        for (int i=0; i<M; i++)
            lambda[i] = (i+1.0) * branchOrders[i];
        setSource("linear");
    }
    else if (initial == IGRandom) {
        double alsum = 0.0;
        for (int i=0; i<N; i++)
            alsum += alpha[i] = rand();
        for (int i=0; i<N; i++)
            alpha[i] /= alsum;

        for (int i=0; i<M; i++)
            lambda[i] = branchOrders[i] / trace->getIATime(rand() % trace->size());
        setSource("random");
    }
    else if (initial == IGExponential) {
        for (int i=0; i<N; i++)
            alpha[i] = 1.0 / N;
        for (int i=0; i<M; i++)
            lambda[i] = pow(10.0, i) * branchOrders[i];
        setSource("exponential");
    }
    // initial value for P
    for (int c=0; c<C; c++)
        for (int i=0; i<M; i++)
            for (int j=0; j<N; j++)
                P[c*N*M+i*N+j] = alpha[j]/C;
    updateBigP();

    vector<double> moms = getMoments(1);
    scaleIntensities (moms[1]/trMean);
}

std::vector<double> SpecMAP::getBranchMoments(int order) const {

    vector<double> moms (N);
    int br = 0;
    int ix = 0;
    for (int i=0; i<N; i++) {
        moms[i] = 1.0 / pow (lambda[br], order);
        for (int j=0; j<order; j++)
            moms[i] *= branchOrders[br]+j-ix;
        ix++;
        if (ix>=branchOrders[br]) {
            br++;
            ix = 0;
        }
    }
    return moms;
}

void SpecMAP::printSummary(ostream &os) const {

    os << "LogLi = " << logli << " (";
    for (int i=0; i<M; i++)
        os << branchOrders[i] << ", ";
    os << source << ")" << endl;
}

void SpecMAP::printAsPH (std::ostream& os, const std::string& format) const {

    vector<vector<double> > D0 (N);
    for (int i=0; i<N; i++)
        D0[i].resize(N);

    // Matrix D0
    // ---------
    // fill with zero
    for_each(D0.begin(), D0.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0;
    for (int i=0; i<M; i++)
        for (int b=0; b<branchOrders[i]; b++) {
            D0[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                D0[ix][ix+1] = lambda[i];
            ix++;
        }

    vector<double> alp (alpha, alpha+N);

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        os << "A = ";
        printMatrix (os, D0, format);
        os << endl << "alpha = ";
        printVector (os, alp, format);
        os << endl;
    }
    else if (format=="numpy") {
        os << "A = matrix(";
        printMatrix(os, D0, format);
        os << ")" << endl << "alpha = matrix([";
        printVector (os, alp, format);
        os << "])" << endl;
    }
}

void SpecMAP::printAsMAP(ostream &os, const string &format) const {

    vector<vector<vector<double> > > D;

    vector<vector<double> > mat;
    mat.resize(N);
    for (int i=0; i<N; i++)
        mat[i].resize(N);

    // Matrix D0
    // ---------
    // fill with zero
    for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
    // fill with hyper-Erlang
    int ix = 0;
    for (int i=0; i<M; i++)
        for (int b=0; b<branchOrders[i]; b++) {
            mat[ix][ix] = - lambda[i];
            if (b<branchOrders[i]-1)
                mat[ix][ix+1] = lambda[i];
            ix++;
        }
    D.push_back(mat);

    // Matrix Dk
    // ---------
    vector<int> endState;
    endState.push_back(branchOrders[0]-1);
    for (int i=1; i<M; i++)
        endState.push_back(endState.back() + branchOrders[i]);

    for (int c=0; c<C; c++) {
        // fill with zero
        for_each(mat.begin(), mat.end(), [](vector<double>& m) { fill(m.begin(), m.end(), 0.0); });
        ix = 0;
        for (int i=0; i<M; i++)
            for (int j=0; j<N; j++)
                mat[endState[i]][j] = lambda[i] * P[c*M*N+i*N+j];
        D.push_back(mat);
    }

    // Print to the stream in the desired format
    if (format=="text" || format=="matlab" || format=="mathematica") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = ";
            printMatrix(os, D[k], format);
            os << endl;
        }
    }
    else if (format=="numpy") {
        for (int k=0; k<D.size(); k++) {
            os << "D" << k << " = matrix(";
            printMatrix(os, D[k], format);
            os << ")" << endl;
        }
    }
}

void SpecMAP::scaleIntensities(double sc) {

    for (int i=0; i<M; i++)
        lambda[i] *= sc;
}

void SpecMAP::updateBigP() {

    delete [] bigP;
    bigP = new double[C*N*N];
    for (int c=0; c<C; c++) {
        int br = 0;
        int ix = 0;
        for (int i=0; i<N; i++) {
            for (int j=0; j<N; j++) {
                bigP[c*N*N+i*N+j] = P[c*N*M+br*N+j];
            }
            ix++;
            if (ix>=branchOrders[br]) {
                br++;
                ix = 0;
            }
        }
    }
}
