//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "spemfit.h"
#include <algorithm>
#include <cstring>
#include <omp.h>
#include <set>
#include <limits>

using namespace std;

SPEMFit::SPEMFit(Trace *tr) : trace(tr), strategy(All) {

    trMean = trace->getMoments(1)[1];
}

void SPEMFit::initFitting(std::vector<int> brOrders, Structure stru, InitialGuess initial) {

    this->stru = stru;

    if (trace->hasZeroIATime() && !checkForOrder1(brOrders)) {
        cerr << "Zero data found, and there is no order-1 branch. Likelihood is zero." << endl;
        exit (1);
    }

    if (initial==IGCombined) {
        initFitting (brOrders, stru, IGLinear);
        initFitting (brOrders, stru, IGExponential);
        initFitting (brOrders, stru, IGKMeans);
        initFitting (brOrders, stru, IGRandom);
        initFitting (brOrders, stru, IGRandom);
        initFitting (brOrders, stru, IGRandom);
    }

    sort (brOrders.begin(), brOrders.end());

    MixtureDensityMAP* map;
    if (stru==StHErMAP)
        map = new HErMAP (brOrders, trace->getNumOfTypes());
    else if (stru==StSpecMAP)
        map = new SpecMAP (brOrders, trace->getNumOfTypes());
    else if (stru==StFEBMAP)
        map = new FEBMAP (brOrders, trace->getNumOfTypes());

    map->setInitialGuess(initial, trace, trMean);
    maps.push_back(map);
}

void SPEMFit::initFitting (int totalStates, BranchOrders brOrders, Structure stru, InitialGuess initial) {

    strategy = brOrders;
    this->totalStates = totalStates;
    this->stru = stru;
    this->initial = initial;

    if (strategy==Incremental)
        return;

    if (initial==IGCombined) {
        initFitting (totalStates, brOrders, stru, IGLinear);
        initFitting (totalStates, brOrders, stru, IGExponential);
        initFitting (totalStates, brOrders, stru, IGKMeans);
        initFitting (totalStates, brOrders, stru, IGRandom);
        initFitting (totalStates, brOrders, stru, IGRandom);
        initFitting (totalStates, brOrders, stru, IGRandom);
    }

    for (int br=2; br<=totalStates; br++) {
        list<vector<int> > l = allOrders (br, totalStates);
        for (list<vector<int> >::iterator i = l.begin(); i!=l.end(); ++i) {
            if (!trace->hasZeroIATime() || checkForOrder1(*i))
                initFitting (*i, stru, initial);
        }
    }
}

bool SPEMFit::checkForOrder1(const std::vector<int> &brOrders) {

    for (int i=0; i<brOrders.size(); i++)
        if (brOrders[i]==1)
            return true;
    return false;
}

std::list<std::vector<int> > SPEMFit::allOrders(int branches, int sumOrders) {

    if (branches==1) {
        vector<int> v;
        v.push_back(sumOrders);
        list<vector<int> > l;
        l.push_back(v);
        return l;
    }
    else {
        list<vector<int> > o;
        for (int i=0; i<=sumOrders-branches; i++) {
            list<vector<int> > x = allOrders (branches-1, sumOrders-i-1);
            for (list<vector<int> >::iterator j=x.begin(); j!=x.end(); j++) {
                j->push_back(i+1);
                sort(j->begin(), j->end());
                // check if we have it already
                bool found = 0;
                for (list<vector<int> >::iterator k=o.begin(); k!=o.end(); k++)
                    if (equal(k->begin(), k->end(), j->begin())) {
                        found = true;
                        break;
                    }
                if (!found)
                    o.push_back(*j);
            }
        }
        return o;
    }
}

void SPEMFit::genNeighbours(const std::vector<int> &brOrders) {

    maps.clear();
    set<vector<int> > allNeigh;
    for (int i=0; i<brOrders.size(); i++) {
        vector<int> newBr = brOrders;
        newBr[i]++;
        sort(newBr.begin(), newBr.end());
        // check if we have it already
        if (allNeigh.find(newBr)==allNeigh.end())
            allNeigh.insert(newBr);
    }
    for (set<vector<int> >::iterator i = allNeigh.begin(); i!=allNeigh.end(); i++) {
        vector<int> bro = *i;
        while (bro.front()==0)
            bro.erase (bro.begin());
        if (!trace->hasZeroIATime() || checkForOrder1(*i))
            initFitting(bro, stru, initial);
    }
}

double SPEMFit::refine(MixtureDensityMAP *ini, bool ph, bool da, bool verbose, int maxiter, double precision) {

    vector<double> temps;
    if (da) {
        double beta_min = 0.5;
        double beta_max = 1.5;
        int upsteps = 11;
        int downsteps = 5;
        for (int i=0; i<upsteps; i++)
            temps.push_back(beta_min + i * (beta_max-beta_min) / (upsteps-1));
        for (int i=0; i<downsteps; i++)
            temps.push_back(beta_max - i * (beta_max - 1.0) / (downsteps-1));
    }
    else
        temps.push_back(1.0);

    for (int i=0; i<temps.size(); i++) {
//        cerr << "Step " << i+1 << ", beta=" << temps[i] << endl;
        int maxi = fabs(temps[i]-1.0) < 1e-9 ? ini->iter+maxiter : ini->iter+10;
        if (stru == StHErMAP) {
            HErMAP* her = dynamic_cast<HErMAP*>(ini);
            if (ph)
                refinePH (her, verbose, maxi, precision, temps[i]);
            else
                refineMMAP (her, verbose, maxi, precision, temps[i]);
        }
        else if (stru == StSpecMAP) {
            SpecMAP* spec = dynamic_cast<SpecMAP*>(ini);
            if (ph)
                refinePH (spec, verbose, maxi, precision, temps[i]);
            else
                refineMMAP (spec, verbose, maxi, precision, temps[i]);
        }
        else if (stru == StFEBMAP) {
            FEBMAP* febm = dynamic_cast<FEBMAP*>(ini);
            if (ph)
                refinePH (febm, verbose, maxi, precision, temps[i]);
        }
    }
    return ini->logli;
}

bool IsNan( float value ) { return ((*(unsigned*)&value) & 0x7fffffff) > 0x7f800000; }
bool MixtureDensityMAPCmp (MixtureDensityMAP* a, MixtureDensityMAP* b) {
    if (IsNan((float)a->logli))
        return false;
    else if (IsNan((float)b->logli))
        return true;
    else
        return a->logli > b->logli;
}

MixtureDensityMAP* SPEMFit::runFitting(bool ph, bool da, double eps, int maxIter, bool verbose) {

    if (strategy==All) {
        #pragma omp parallel for if(maps.size()>1) schedule(dynamic)
        for (int k=0; k<maps.size(); k++) {
            refine (maps[k], ph, da, maps.size()==1, maxIter, eps);
            maps[k]->printSummary (cout);
        }
        sort (maps.begin(), maps.end(), MixtureDensityMAPCmp);
    }
    else if (strategy==Incremental) {
        vector<int> bestBrOrders (totalStates, 0);
        MixtureDensityMAP* bestMAP = NULL;
        bestBrOrders[totalStates-1] = 1.0;
        for (int i=1; i<totalStates; i++) {
            // find all neighbours of bestBrOrders, and construct a MAP for each
            genNeighbours(bestBrOrders);
            // try them all and select the best
//            #pragma omp parallel for if(maps.size()>1) schedule(dynamic)
            #pragma omp for
            for (int k=0; k<maps.size(); k++) {
                refine (maps[k], ph, da, maps.size()==1, maxIter, eps);
                maps[k]->printSummary (cout);
            }
            // select the best one and store the corresponding brach orders
            sort (maps.begin(), maps.end(), MixtureDensityMAPCmp);
            fill (bestBrOrders.begin(), bestBrOrders.end(), 0);
            copy (maps.front()->getErlangOrders(), maps.front()->getErlangOrders() + maps.front()->getNumOfErlangs(), bestBrOrders.begin());
            if (verbose) {
                cout << "=== best: ";
                maps.front()->printSummary (cout);
            }
        }
    }
    else if (strategy==ProgPres) {
        vector<double> prec;
        double minPrec = 5e-4;
        double maxPrec = eps * 2.0;
        int steps = 4;
        for (int i=0; i<steps; i++)
            prec.push_back(exp( log(maxPrec) + (steps-1-i)*(log(minPrec)-log(maxPrec))/(steps-1.0) ));

        int start = max(0, (int)prec.size() - (int)floor(log2(maps.size())));

    //    prec.push_back(5e-4);
    //    prec.push_back(1e-4);
    //    prec.push_back(5e-6);
    //    prec.push_back(1e-6);

        for (int j=start; j<prec.size() && maps.size()>=2; j++) {
            cout << "Progressive selection round " << j+1 << " (prec=" << prec[j] << "):" << endl;
            #pragma omp parallel for if(maps.size()>1) schedule(dynamic)
            for (int k=0; k<maps.size(); k++) {
                refine (maps[k], ph, da, false, maxIter, prec[j]);
                maps[k]->printSummary (cout);
            }
            sort (maps.begin(), maps.end(), MixtureDensityMAPCmp);
            if (maps.size() >= 2)
                maps.erase(maps.begin()+maps.size()/2, maps.end());
        }
        refine (maps.front(), ph, false, verbose, maxIter, eps);
    }

    return maps.front();
}

double SPEMFit::refinePH (FEBMAP *ini, bool verbose, int maxiter, double precision, double beta) {

    int M = ini->M;
    int K = trace->size();
    int T = omp_get_max_threads();

    double* sumqmkt = new double [T*M];
    double* sumxqmkt = new double [T*M];
    double* sumkqmkt = new double [T*M];
    double* loglit  = new double [T];
    double* sqmk = new double [T*M];
    double* skqmk = new double [T*M];

    double diff = 1.0, logli, ologli = - numeric_limits<double>::infinity();

    int iter = ini->iter;
    bool stop = false;
    while (iter < maxiter && diff > precision) {

        memset(sumqmkt, 0, T*M*sizeof(double));
        memset(sumxqmkt, 0, T*M*sizeof(double));
        memset(sumkqmkt, 0, T*M*sizeof(double));
        memset(loglit, 0, T*sizeof(double));
        #pragma omp parallel num_threads(T)
        {
            int tid = omp_get_thread_num();
            int tnum = omp_get_num_threads();
            for (int k=tid; k<K; k+=tnum) {
                double x = trace->getIATime(k);

                double sum = 0.0;
                for (int m=0; m<M; m++) {
                    double fx = (1.0 - ini->z[m]) * ini->lambda[m] * exp(-ini->lambda[m]*x);
                    double n = 1.0;
                    for (int l=1; l<ini->branchOrders[m]; l++, n+=1.0)
                        fx *= ini->lambda[m] * x / n;
                    int kk = 1;
                    sqmk[tid*M+m] = skqmk[tid*M+m] = fx;
                    while (fx > 1e-14) {
                        kk++;
                        for (int l=0; l<ini->branchOrders[m]; l++, n+=1.0)
                            fx *= ini->lambda[m] * x / n;
                        fx *= ini->z[m];
                        sqmk[tid*M+m] += fx;
                        skqmk[tid*M+m] += kk * fx;
//                        cerr << fx << endl;
                    }
                    sum += sqmk[tid*M+m] *= ini->alpha[m];
                    skqmk[tid*M+m] *= ini->alpha[m];
                }

                if (sum==0) {
                    stop = true;
                    break;
                }
                for (int m=0; m<M; m++) {
                    sqmk[tid*M+m] /= sum;
                    skqmk[tid*M+m] /= sum;
                    sumqmkt[tid*M+m] += sqmk[tid*M+m];
                    sumxqmkt[tid*M+m] += x * sqmk[tid*M+m];
                    sumkqmkt[tid*M+m] += skqmk[tid*M+m];
                }
                loglit[tid] += log(sum);
            }
        }
        if (stop) {
            delete [] sumqmkt;
            delete [] sumxqmkt;
            delete [] sumkqmkt;
            delete [] loglit;
            delete [] sqmk;
            delete [] skqmk;
            return - numeric_limits<double>::infinity();
        }

        logli = 0.0;
        for (int t=0; t<T; t++)
            logli += loglit[t];

        logli = logli / K;
        for (int m=0; m<M; m++) {
            double sumqmk = 0.0, sumxqmk = 0.0, sumkqmk = 0.0;
            for (int t=0; t<T; t++) {
                sumqmk  += sumqmkt[t*M+m];
                sumxqmk += sumxqmkt[t*M+m];
                sumkqmk += sumkqmkt[t*M+m];
            }
            ini->alpha[m]  = sumqmk / K;
            ini->lambda[m] = ini->branchOrders[m] * sumkqmk / sumxqmk;
            ini->z[m]      = 1.0 - sumqmk / sumkqmk;
//            cerr << ini->lambda[m] << ", " << ini->z[m] << ", " << ini->alpha[m] << endl;
        }
        diff = fabs((logli - ologli) / logli);
        ologli = logli;
        iter++;
        if (verbose)
            std::cout << "Iteration " << iter << ": diff = " << diff << ", logli = " << logli << std::endl;

        ini->logli = logli;
        ini->iter = iter;
    }

    delete [] sumqmkt;
    delete [] sumxqmkt;
    delete [] sumkqmkt;
    delete [] loglit;
    delete [] sqmk;
    delete [] skqmk;
    return logli;
}

double SPEMFit::refinePH (HErMAP *ini, bool verbose, int maxiter, double precision, double beta) {

    int M = ini->M;
    int K = trace->size();
    int T = omp_get_max_threads();

    double* sumqkt = new double [T*M];
    double* sumxqkt = new double [T*M];
    double* loglit  = new double [T];
    double* qk = new double [T*M];

    double diff = 1.0, logli, ologli = - numeric_limits<double>::infinity();

    int iter = ini->iter;
    bool stop = false;
    while (iter < maxiter && diff > precision) {

        memset(sumqkt, 0, T*M*sizeof(double));
        memset(sumxqkt, 0, T*M*sizeof(double));
        memset(loglit, 0, T*sizeof(double));
        #pragma omp parallel num_threads(T)
        {
            int tid = omp_get_thread_num();
            int tnum = omp_get_num_threads();
            for (int k=tid; k<K; k+=tnum) {
                double x = trace->getIATime(k);
                double sum = 0.0;
                for (int m=0; m<M; m++) {
                    qk[tid*M+m] = exp(beta*log(ini->alpha[m]*ini->getBranchDensity (x, m)));
                    sum += qk[tid*M+m];
                }
                if (sum==0) {
                    stop = true;
                    break;
                }
                for (int m=0; m<M; m++) {
                    qk[tid*M+m] /= sum;
                    sumqkt[tid*M+m] += qk[tid*M+m];
                    sumxqkt[tid*M+m] += x * qk[tid*M+m];
                }
                loglit[tid] += log(sum);
            }
        }
        if (stop) {
            delete [] sumqkt;
            delete [] sumxqkt;
            delete [] loglit;
            delete [] qk;
            return - numeric_limits<double>::infinity();
        }

        logli = 0.0;
        for (int t=0; t<T; t++)
            logli += loglit[t];

        logli = logli / K;
        for (int m=0; m<M; m++) {
            double sumqk = 0.0, sumxqk = 0.0;
            for (int t=0; t<T; t++) {
                sumqk  += sumqkt[t*M+m];
                sumxqk += sumxqkt[t*M+m];
            }
            ini->alpha[m] = sumqk / K;
            ini->lambda[m] = ini->branchOrders[m] * sumqk / sumxqk;
        }
        diff = fabs((logli - ologli) / logli);
        ologli = logli;
        iter++;
        if (verbose)
            std::cout << "Iteration " << iter << ": diff = " << diff << ", logli = " << logli << std::endl;

        ini->logli = logli;
        ini->iter = iter;
    }

    delete [] sumqkt;
    delete [] sumxqkt;
    delete [] loglit;
    delete [] qk;
    return logli;
}


double SPEMFit::refineMMAP (HErMAP *ini, bool verbose, int maxiter, double precision, double beta) {

    int M = ini->M;
    int K = trace->size();
    int C = trace->getNumOfTypes();
    int T = omp_get_max_threads();

    double* sumqk = new double [T*M];
    double* sumxqk = new double [T*M];

    double* fMat = new double[M*K];
    double* aMat = new double[M*K];
    double* bMat = new double[M*K];
    int* aScale = new int[K];
    int* bScale = new int[K];

    double diff = 1.0, logli, ologli = - numeric_limits<double>::infinity();

    double* newP = new double[T*C*M*M];
    double* qVecCurr = new double[T*M];

    int iter = ini->iter;
    double log2 = log(2.0);
    bool stop = false;

    while (iter < maxiter && diff > precision) {

        // Calculate vector f (branch densities) for each data. We can do it in parallel.
        #pragma omp parallel for
        for (int k=0; k<K; k++) {
            double x = trace->getIATime(k);
            for (int m=0; m<M; m++)
                fMat[k*M+m] = ini->getBranchDensity(x, m);
        }

        // calculate vector a and vector b (forward and backward likelihood vectors)
        memset(aMat, 0, M*K*sizeof(double));
        memset(bMat, 0, M*K*sizeof(double));
        memset(aScale, 0, K*sizeof(int));
        memset(bScale, 0, K*sizeof(int));

        #pragma omp parallel num_threads(2)
        {
            if (omp_get_thread_num()==0) {
                // The first thread computes the forward likelihood vectors (a)
                // 0th element
                for (int i=0; i<M; i++)
                    for (int j=0; j<M; j++)
                        aMat[i] += ini->alpha[j]*fMat[j]*ini->P[trace->getType(0)*M*M+j*M+i];
                // all other elements
                for (int k=1; k<K; k++) {
                    int tk = trace->getType(k);
                    int ofs = k*M;
                    double asum = 0.0;
                    for (int i=0; i<M; i++) {
                        for (int j=0; j<M; j++)
                            aMat[ofs+i] += aMat[ofs+j-M]*fMat[ofs+j]*ini->P[tk*M*M+j*M+i];
                        asum += aMat[ofs+i];
                    }
                    aScale[k] = aScale[k-1];
                    if (asum==0) {
                        stop = true;
                        break;
                    }
                    while (asum>1.0) {
                        aScale[k]++;
                        asum /= 2.0;
                        for (int i=0; i<M; i++)
                            aMat[ofs+i] /= 2.0;
                    }
                    while (asum<0.1) {
                        aScale[k]--;
                        asum *= 2.0;
                        for (int i=0; i<M; i++)
                            aMat[ofs+i] *= 2.0;
                    }
                }
            }
            if (omp_get_num_threads()==1 || omp_get_thread_num()>0) {
                // The second thread computes the backward likelihood vectors (b)
                // (k-1)th element
                int ofs = (K-1)*M;
                for (int j=0; j<M; j++)
                    for (int i=0; i<M; i++)
                        bMat[ofs+j] += fMat[ofs+j]*ini->P[trace->getType(K-1)*M*M+j*M+i];
                // all other elements
                for (int k=K-2; k>=0; k--) {
                    int tk = trace->getType(k);
                    ofs = k*M;
                    double bsum = 0.0;
                    for (int j=0; j<M; j++) {
                        for (int i=0; i<M; i++)
                            bMat[ofs+j] += bMat[ofs+i+M]*fMat[ofs+j]*ini->P[tk*M*M+j*M+i];
                        bsum += bMat[ofs+j];
                    }
                    bScale[k] = bScale[k+1];
                    if (bsum==0) {
                        stop = true;
                        break;
                    }
                    while (bsum>1.0) {
                        bScale[k]++;
                        bsum /= 2.0;
                        for (int j=0; j<M; j++)
                            bMat[ofs+j] /= 2.0;
                    }
                    while (bsum<0.1) {
                        bScale[k]--;
                        bsum *= 2.0;
                        for (int j=0; j<M; j++)
                            bMat[ofs+j] *= 2.0;
                    }
                }
            }
        }

        if (stop) {
            delete [] aMat;
            delete [] bMat;
            delete [] fMat;
            delete [] aScale;
            delete [] bScale;
            delete [] newP;
            delete [] qVecCurr;
            delete [] sumqk;
            delete [] sumxqk;
            return - numeric_limits<double>::infinity();
        }

        // Calculate log-likelihood
        double llh = 0.0;
        for (int i=0; i<M; i++)
            llh += ini->alpha[i]*bMat[i];
        logli = (log(llh) + bScale[0] * log2) / K;

        // The inverse of log-likelihood
        double illh = 1.0 / llh;

        // Calculate component means and the new state transition matrix
        memset(newP, 0, T*C*M*M*sizeof(double));
        memset(sumqk, 0, T*M*sizeof(double));
        memset(sumxqk, 0, T*M*sizeof(double));

        // Calculate new estimates for the parameters, in parallel
        int tNum;
        #pragma omp parallel
        {
            int tnum = omp_get_num_threads();
            int tid = omp_get_thread_num();

            for (int k=tid; k<K; k+=tnum) {
                int tk = trace->getType(k);
                int ofs = k*M;
                double x = trace->getIATime(k);
                double qVecSum = 0.0;
                double normalizer;
                if (k==0)
                    normalizer = ldexp (illh, bScale[1] - bScale[0]);
                else if (k<K-1)
                    normalizer = ldexp(illh, aScale[k-1] + bScale[k+1] - bScale[0]);
                else
                    normalizer = ldexp(illh, aScale[k-1] - bScale[0]);

                for (int m=0; m<M; m++) {
                    double pMul = (k==0 ? ini->alpha[m] : aMat[ofs+m-M]);
                    qVecSum += qVecCurr[tid*M + m] = pMul * bMat[ofs+m];
                    pMul *= fMat[ofs+m] * normalizer;
                    if (k<K-1) {
                        for (int j=0; j<M; j++)
                            newP[tid*C*M*M+tk*M*M+m*M+j] += pMul * ini->P[tk*M*M+m*M+j] * bMat[ofs+M+j];
                    }
                    else
                        for (int j=0; j<M; j++)
                            newP[tid*C*M*M+tk*M*M+m*M+j] += pMul * ini->P[tk*M*M+m*M+j] * 1;
                }
                for (int m=0; m<M; m++) {
                    sumqk[tid*M+m] += qVecCurr[tid*M+m] / qVecSum;
                    sumxqk[tid*M+m] += x* qVecCurr[tid*M+m] / qVecSum;
                }
            }
            if (tid==0)
                tNum = tnum;
        }

        // Collect the results of the threads
        for (int t=1; t<tNum; t++) {
            for (int m=0; m<M; m++) {
                sumqk[m] += sumqk[t*M+m];
                sumxqk[m] += sumxqk[t*M+m];
            }
            for (int m=0; m<C*M*M; m++)
                newP[m] += newP[t*C*M*M+m];
        }

        // Store new estimates
        for (int i=0; i<M; i++) {
            ini->lambda[i] = ini->branchOrders[i] * sumqk[i] / sumxqk[i];
            ini->alpha[i] = sumqk[i] / K;
            double rowSum = 0.0;
            for (int c=0; c<C; c++)
                for (int j=0; j<M; j++)
                    rowSum += newP[c*M*M+i*M+j];
            for (int c=0; c<C; c++)
                for (int j=0; j<M; j++)
                    ini->P[c*M*M+i*M+j] = newP[c*M*M+i*M+j] / rowSum;
        }

        // Calculate relative difference of the log likelihoods
        diff = fabs((logli - ologli) / logli);
        ologli = logli;
        iter++;
        if (verbose)
            std::cout << "Iteration " << iter << ": diff = " << diff << ", logli = " << logli << std::endl;

        ini->logli = logli;
        ini->iter = iter;
    }

    delete [] aMat;
    delete [] bMat;
    delete [] fMat;
    delete [] aScale;
    delete [] bScale;
    delete [] newP;
    delete [] qVecCurr;
    delete [] sumqk;
    delete [] sumxqk;

    return logli;
}

double SPEMFit::refinePH (SpecMAP *ini, bool verbose, int maxiter, double precision, double beta) {

    int N = ini->N;
    int M = ini->M;
    int K = trace->size();
    int T = omp_get_max_threads();

    double* sumqkt = new double [T*N];
    double* sumxqkt = new double [T*N];
    double* loglit  = new double [T];
    double* qk = new double [T*N];

    double diff = 1.0, logli, ologli = - numeric_limits<double>::infinity();

    int iter = ini->iter;
    bool stop = false;
    while (iter < maxiter && diff > precision) {

        memset(sumqkt, 0, T*N*sizeof(double));
        memset(sumxqkt, 0, T*N*sizeof(double));
        memset(loglit, 0, T*sizeof(double));
        int tNum;
        #pragma omp parallel num_threads(T)
        {
            int tid = omp_get_thread_num();
            int tnum = omp_get_num_threads();
            for (int k=tid; k<K; k+=tnum) {
                double x = trace->getIATime(k);
                int n = N-1;
                for (int m=M-1; m>=0; m--) {
                    qk[tid*N+n] = ini->lambda[m] * exp(-ini->lambda[m] * x);
                    n--;
                    for (int e=1; e<ini->branchOrders[m]; e++) {
                        qk[tid*N+n] = qk[tid*N+n+1] * ini->lambda[m] * x / e;
                        n--;
                    }
                }
                double sum = 0.0;
                for (int m=0; m<N; m++) {
                    qk[tid*N+m] = exp(beta*log(ini->alpha[m]*qk[tid*N+m]));
                    sum += qk[tid*N+m];
                }
                if (sum==0) {
                    stop = true;
                    break;
                }
                for (int m=0; m<N; m++) {
                    qk[tid*N+m] /= sum;
                    sumqkt[tid*N+m] += qk[tid*N+m];
                    sumxqkt[tid*N+m] += x * qk[tid*N+m];
                }
                loglit[tid] += log(sum);
            }
            if (tid==0)
                tNum = tnum;
        }
        if (stop) {
            delete [] sumqkt;
            delete [] sumxqkt;
            delete [] loglit;
            delete [] qk;
            return - numeric_limits<double>::infinity();
        }
        logli = 0.0;
        for (int t=0; t<tNum; t++)
            logli += loglit[t];

        logli = logli / K;

        for (int t=1; t<tNum; t++)
            for (int m=0; m<N; m++) {
                sumqkt[m] += sumqkt[t*N+m];
                sumxqkt[m] += sumxqkt[t*N+m];
            }
        // Store new estimates
        // alpha
        for (int i=0; i<N; i++)
            ini->alpha[i] = sumqkt[i] / K;
        // lambda
        int br = 0;
        int ix = ini->branchOrders[br];
        double norm = 0.0;
        memset(ini->lambda, 0, M*sizeof(double));
        for (int i=0; i<N; i++) {
            ini->lambda[br] += ix * sumqkt[i];
            norm += sumxqkt[i];
            ix--;
            if (ix==0) {
                ini->lambda[br] /= norm;
                br++;
                norm = 0.0;
                if (br<M)
                    ix = ini->branchOrders[br];
            }
        }
        diff = fabs((logli - ologli) / logli);
        ologli = logli;
        iter++;
        if (verbose)
            std::cout << "Iteration " << iter << ": diff = " << diff << ", logli = " << logli << std::endl;

        ini->logli = logli;
        ini->iter = iter;
    }

    delete [] sumqkt;
    delete [] sumxqkt;
    delete [] loglit;
    delete [] qk;
    return logli;
}

double SPEMFit::refineMMAP (SpecMAP *ini, bool verbose, int maxiter, double precision, double beta) {

    int M = ini->M;
    int N = ini->N;
    int K = trace->size();
    int C = trace->getNumOfTypes();
    int T = omp_get_max_threads();

    double* sumqk = new double [T*N];
    double* sumxqk = new double [T*N];

    double* fMat = new double[N*K];
    double* aMat = new double[N*K];
    double* bMat = new double[N*K];
    int* aScale = new int[K];
    int* bScale = new int[K];

    double diff = 1.0, logli, ologli = - numeric_limits<double>::infinity();

    double* newP = new double[T*C*M*N];
    double* qVecCurr = new double[T*N];

    int iter = ini->iter;
    double log2 = log(2.0);
    bool stop = false;

    while (iter < maxiter && diff > precision) {

        // Calculate vector f (branch densities) for each data. We can do it in parallel.
        #pragma omp parallel for
        for (int k=0; k<K; k++) {
            double x = trace->getIATime(k);
            int n = N-1;
            for (int m=M-1; m>=0; m--) {
                fMat[k*N+n] = ini->lambda[m] * exp(-ini->lambda[m] * x);
                n--;
                for (int e=1; e<ini->branchOrders[m]; e++) {
                    fMat[k*N+n] = fMat[k*N+n+1] * ini->lambda[m] * x / e;
                    n--;
                }
            }
        }

        // calculate vector a and vector b (forward and backward likelihood vectors)
        memset(aMat, 0, N*K*sizeof(double));
        memset(bMat, 0, N*K*sizeof(double));
        memset(aScale, 0, K*sizeof(int));
        memset(bScale, 0, K*sizeof(int));

        #pragma omp parallel num_threads(2)
        {
            if (omp_get_thread_num()==0) {
                // The first thread computes the forward likelihood vectors (a)
                // 0th element
                for (int i=0; i<N; i++)
                    for (int j=0; j<N; j++)
                        aMat[i] += ini->alpha[j]*fMat[j]*ini->bigP[trace->getType(0)*N*N+j*N+i];
                // all other elements
                for (int k=1; k<K; k++) {
                    int tk = trace->getType(k);
                    int ofs = k*N;
                    double asum = 0.0;
                    for (int i=0; i<N; i++)
                        for (int j=0; j<N; j++)
                            asum += aMat[ofs+i] += aMat[ofs+j-N]*fMat[ofs+j]*ini->bigP[tk*N*N+j*N+i];
                    aScale[k] = aScale[k-1];
                    if (asum==0) {
                        stop = true;
                        cerr << "Bad news! " << k << endl;
                        break;
                    }
                    while (asum>1.0) {
                        aScale[k]++;
                        asum /= 2.0;
                        for (int i=0; i<N; i++)
                            aMat[ofs+i] /= 2.0;
                    }
                    while (asum<0.1) {
                        aScale[k]--;
                        asum *= 2.0;
                        for (int i=0; i<N; i++)
                            aMat[ofs+i] *= 2.0;
                    }
                }
            }
            if (omp_get_num_threads()==1 || omp_get_thread_num()>0) {
                // The second thread computes the backward likelihood vectors (b)
                // (k-1)th element
                int ofs = (K-1)*N;
                for (int j=0; j<N; j++)
                    for (int i=0; i<N; i++)
                        bMat[ofs+j] += fMat[ofs+j]*ini->bigP[trace->getType(K-1)*N*N+j*N+i];
                // all other elements
                for (int k=K-2; k>=0; k--) {
                    int tk = trace->getType(k);
                    ofs = k*N;
                    double bsum = 0.0;
                    for (int j=0; j<N; j++)
                        for (int i=0; i<N; i++)
                            bsum += bMat[ofs+j] += bMat[ofs+i+N]*fMat[ofs+j]*ini->bigP[tk*N*N+j*N+i];
                    bScale[k] = bScale[k+1];
                    if (bsum==0) {
                        stop = true;
                        break;
                    }
                    while (bsum>1.0) {
                        bScale[k]++;
                        bsum /= 2.0;
                        for (int j=0; j<N; j++)
                            bMat[ofs+j] /= 2.0;
                    }
                    while (bsum<0.1) {
                        bScale[k]--;
                        bsum *= 2.0;
                        for (int j=0; j<N; j++)
                            bMat[ofs+j] *= 2.0;
                    }
                }
            }
        }
        if (stop) {
            delete [] aMat;
            delete [] bMat;
            delete [] fMat;
            delete [] aScale;
            delete [] bScale;
            delete [] newP;
            delete [] qVecCurr;
            delete [] sumqk;
            delete [] sumxqk;
            return - numeric_limits<double>::infinity();
        }

        // Calculate log-likelihood
        double llh = 0.0;
        for (int i=0; i<N; i++)
            llh += ini->alpha[i]*bMat[i];
        logli = (log(llh) + bScale[0] * log2) / K;

        // The inverse of log-likelihood
        double illh = 1.0 / llh;

        // Calculate component means and the new state transition matrix
        memset(newP, 0, T*C*N*M*sizeof(double));
        memset(sumqk, 0, T*N*sizeof(double));
        memset(sumxqk, 0, T*N*sizeof(double));

        // Calculate new estimates for the parameters, in parallel
        int tNum;
        #pragma omp parallel
        {
            int tnum = omp_get_num_threads();
            int tid = omp_get_thread_num();

            for (int k=tid; k<K; k+=tnum) {
                int tk = trace->getType(k);
                int ofs = k*N;
                double x = trace->getIATime(k);
                double qVecSum = 0.0;
                double normalizer;
                if (k==0)
                    normalizer = ldexp (illh, bScale[1] - bScale[0]);
                else if (k<K-1)
                    normalizer = ldexp(illh, aScale[k-1] + bScale[k+1] - bScale[0]);

                int br = 0;
                int ix = 0;
                for (int m=0; m<N; m++) {
                    double pMul = (k==0 ? ini->alpha[m] : aMat[ofs+m-N]);
                    qVecSum += qVecCurr[tid*N + m] = pMul * bMat[ofs+m];
                    if (k<K-1) {
                        pMul *= fMat[ofs+m] * normalizer;
                        for (int j=0; j<N; j++)
                            newP[tid*C*N*M+tk*N*M+br*N+j] += pMul * ini->bigP[tk*N*N+m*N+j] * bMat[ofs+N+j];
                    }
                    ix++;
                    if (ix>=ini->branchOrders[br]) {
                        br++;
                        ix = 0;
                    }
                }
                for (int m=0; m<N; m++) {
                    sumqk[tid*N+m] += qVecCurr[tid*N+m] / qVecSum;
                    sumxqk[tid*N+m] += x* qVecCurr[tid*N+m] / qVecSum;
                }
            }
            if (tid==0)
                tNum = tnum;
        }
        // Collect the results of the threads
        for (int t=1; t<tNum; t++) {
            for (int m=0; m<N; m++) {
                sumqk[m] += sumqk[t*N+m];
                sumxqk[m] += sumxqk[t*N+m];
            }
            for (int m=0; m<C*N*M; m++)
                newP[m] += newP[t*C*N*M+m];
        }
        // Store new estimates
        // alpha
        for (int i=0; i<N; i++)
            ini->alpha[i] = sumqk[i] / K;
        // lambda
        int br = 0;
        int ix = ini->branchOrders[br];
        double norm = 0.0;
        memset(ini->lambda, 0, M*sizeof(double));
        for (int i=0; i<N; i++) {
            ini->lambda[br] += ix * sumqk[i];
            norm += sumxqk[i];
            ix--;
            if (ix==0) {
                ini->lambda[br] /= norm;
                br++;
                norm = 0.0;
                if (br<M)
                    ix = ini->branchOrders[br];
            }
        }

        // P
        for (int i=0; i<M; i++) {
            double rowSum = 0.0;
            for (int c=0; c<C; c++)
                for (int j=0; j<N; j++)
                    rowSum += newP[c*N*M+i*N+j];
            for (int c=0; c<C; c++)
                for (int j=0; j<N; j++)
                    ini->P[c*N*M+i*N+j] = newP[c*N*M+i*N+j] / rowSum;
        }
        ini->updateBigP();

        // Calculate relative difference of the log likelihoods
        diff = fabs((logli - ologli) / logli);
        ologli = logli;
        iter++;
        if (verbose)
            std::cout << "Iteration " << iter << ": diff = " << diff << ", logli = " << logli << std::endl;

        ini->logli = logli;
        ini->iter = iter;
    }

    delete [] aMat;
    delete [] bMat;
    delete [] fMat;
    delete [] aScale;
    delete [] bScale;
    delete [] newP;
    delete [] qVecCurr;
    delete [] sumqk;
    delete [] sumxqk;

    return logli;
}

