//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <stdlib.h>
#include <chrono>
#include <cmath>
#include "spemfit.h"

using namespace std;

void exitWithMsg (const string& msg) {

    cerr << "Error: " << msg << endl;
    exit (1);
}

void printUsage () {

    cout << endl << "SPEM-Fit, (c) Gabor Horvath, 2013." << endl << endl;
    cout << "This program comes with ABSOLUTELY NO WARRANTY." << endl;
    cout << "This is free software, and you are welcome to redistribute it." << endl;
    cout << "If you use SPEM-Fit, please cite our scientific publication." << endl << endl;

    cout << "Usage: spemfit option1 option2 ..." << endl << endl;
    cout << "Options:" << endl;
    cout << "    -input=<filename>:   specifies the input trace file containing the inter-arrival times [mandatory]" << endl;
    cout << "    -output=<filename>:  specifies the output format of the MAP matrices. Possibilities: text, matlab, mathematica, numpy [optional]" << endl;
    cout << "    -outformat=<format>: specifies the file in which the result is stored [optional], can be \"matlab\",\"mathematica\" or \"numpy\"" << endl;
    cout << "    -multitype:          set this option if multi-type arrivals are fitted [optional, default=no multitype]" << endl;
    cout << "    -ph:                 set this option if you dont want to fit the correlations" << endl;
    cout << "    -verbose:            enables verbose messages [optional, default=false]" << endl;
    cout << "    -evaluate:           performs a short evaluation of the quality of the fitting [optional, default=true]" << endl;
    cout << "    -states=6            the number of states of the resulting MAP." << endl;
    cout << "    -orders=1,2,3        a comma separated list with the order of Erlang branches to be used for fitting [optional]" << endl;
    cout << "    -orders=all          try out all possible branch order configurations up to the given number of states and select the best solution" << endl;
    cout << "    -orders=incr         apply the \"incremental\" method to find the best branch order configuration" << endl;
    cout << "    -orders=prog         apply the \"progressive preselection\" method to find the best branch order configuration" << endl;
    cout << "    -structure=hermap    apply the ER-CHMM structure for MAP/MMAP fitting (equivalent to G-FIT if \"-ph\" option is specified)" << endl;
    cout << "    -structure=specmap   apply a more general structure for MAP/MMAP fitting (default)" << endl;
    cout << "    -structure=febmap    apply the FER-CHMM structure for MAP/MMAP fitting" << endl;
    cout << "    -maxiter=1000        maximum number of iterations [optional, default=1000]" << endl;
    cout << "    -precision=1e-7      the relative amount of change in log-likelihood by which the algorithm stops [optional, default=1e-7]" << endl;
    cout << "    -initial=combined    determines the initial guess from which the iteration is started [optional, default=combined]" << endl;
    cout << "       Possibilities: linear: mean values of the branches increases linearly at the beginning" << endl;
    cout << "                      exponential: mean values of the branches increases exponentially at the beginning" << endl;
    cout << "                      kmeans: mean values of the branches and the branch probabilities are determined by k-means clusterization" << endl;
    cout << "                      random: random initial guess" << endl;
    cout << "                      combined: the fitting is started from all supported kind of initial guesses simultaneously" << endl;
    cout << "    -seed                the random seed. Effective only if \"random\" initial guess has been selected [optional, default=1]" << endl;
    cout << "    -da                  enable deterministic annealing" << endl;
    cout << endl << "Example: \"spemfit -input=pAug89_dat.trace -output=result.txt -states=6 -orders=prog -verbose\"" << endl;

}

vector<string> tokenize(const string& str,const string& delimiters) {

    vector<string> tokens;

    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    string::size_type pos = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos) {
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        lastPos = str.find_first_not_of(delimiters, pos);
        pos = str.find_first_of(delimiters, lastPos);
    }

    return tokens;
}

void message (string str, bool verbose) {

    if (verbose)
        cout << str << endl;
    cout.flush();
}

void evaluateSingleType (Trace* trace, MixtureDensityMAP* hem, bool ph) {

    cout << "Relative errors in the first 10 moments:" << endl;
    vector<double> resMoms = hem->getMoments(10);
    vector<double> traceMoms = trace->getMoments(10);
    for (int i=1; i<resMoms.size(); i++)
        cout << i << "th moment: " << setprecision(3) << 100.0 * fabs(resMoms[i]-traceMoms[i]) / traceMoms[i] << "%" << " (" << resMoms[i] << ", " << traceMoms[i] << ")" << endl;
    if (!ph) {
        cout << "--------------------------" << endl;
        cout << "- Lag-k autocorrelations -" << endl;
        cout << "--------------------------" << endl;
        cout << setprecision(5);
        vector<double> trAcf = trace->getAcf(11);
        vector<double> mapAcf = hem->getAcf(11);
        for (int i=1; i<trAcf.size(); i++)
            cout << mapAcf[i] << "\t" << trAcf[i] << endl;
        cout << "----------- Joint Moment relative errors ------------" << endl;
        vector<vector<double> > trJms = trace->getJointMoments(6);
        vector<vector<double> > mapJms = hem->getJointMoments(6);
        cout << setprecision(2);
        for (int i=1; i<trJms.size(); i++) {
            for (int j=1; j<trJms[i].size(); j++)
                cout << 100.0 * fabs(mapJms[i][j]-trJms[i][j]) / trJms[i][j] << "%\t";
            cout << endl;
        }
    /*    cout << "----------- Joint Moment values (MAP/trace) ------------" << endl;
        cout << setprecision(5);
        for (int i=0; i< trJms.size(); i++) {
            for (int j=0; j< trJms[i].size(); j++)
                cout << mapJms[i][j] << " / " << trJms[i][j] << "\t";
            cout << endl;
        }*/
    }
    cout.flush();
}

void evaluateMultiType (Trace* trace, MixtureDensityMAP* hem, bool ph) {

    evaluateSingleType (trace, hem, ph);
}

int main (int argc, char* argv[]) {

    string input, output, format="matlab";
    bool verbose=false, evaluate=true;
    int states = 6;
    vector<int> branchOrders = {1, 2, 3};
    InitialGuess initial = IGKMeans;
    SPEMFit::BranchOrders strategy = SPEMFit::ProgPres;
    SPEMFit::Structure stru = SPEMFit::StSpecMAP;
    double precision = 1e-7;
    int maxiter = 1000;
    bool useBro = false;
    bool multitype = false;
    bool ph = false;
    bool da = false;

    int ix = 1;
    while (ix<argc) {
        string arg (argv[ix]);
        if (arg.substr(0, 7) == "-input=")
            input = arg.substr(7);
        else if (arg.substr(0, 8) == "-output=")
            output = arg.substr(8);
        else if (arg.substr(0, 8) == "-outformat=")
            format = arg.substr(11);
        else if (arg.substr(0, 8) == "-verbose")
            verbose = true;
        else if (arg.substr(0, 9) == "-evaluate")
            evaluate = true;
        else if (arg.substr(0, 3) == "-ph")
            ph = true;
        else if (arg.substr(0, 3) == "-da")
            da = true;
        else if (arg.substr(0, 10) == "-multitype")
            multitype = true;
        else if (arg.substr(0,9) == "-maxiter=")
            maxiter = atoi (arg.substr(9).c_str());
        else if (arg.substr(0,11) == "-precision=")
            precision = atof (arg.substr(11).c_str());
        else if (arg == "-initial=linear")
            initial = IGLinear;
        else if (arg == "-initial=exponential")
            initial = IGExponential;
        else if (arg == "-initial=kmeans")
            initial = IGKMeans;
        else if (arg == "-initial=random")
            initial = IGRandom;
        else if (arg == "-initial=combined")
            initial = IGCombined;
        else if (arg.substr(0,6) == "-seed=")
            srand (atoi (arg.substr(6).c_str()));
        else if (arg.substr(0,8) == "-states=")
            states = atoi (arg.substr(8).c_str());
        else if (arg.substr(0,11) == "-orders=all")
            strategy = SPEMFit::All;
        else if (arg.substr(0,12) == "-orders=prog")
            strategy = SPEMFit::ProgPres;
        else if (arg.substr(0,12) == "-orders=incr")
            strategy = SPEMFit::Incremental;
        else if (arg.substr(0,8) == "-orders=") {
            useBro = true;
            branchOrders.clear();
            vector<string> orders = tokenize (arg.substr(8), ",");
            for (int j=0; j<orders.size(); j++)
                branchOrders.push_back(atoi(orders[j].c_str()));
        }
        else if (arg.substr(0,18) == "-structure=specmap")
            stru = SPEMFit::StSpecMAP;
        else if (arg.substr(0,17) == "-structure=hermap")
            stru = SPEMFit::StHErMAP;
        else if (arg.substr(0,17) == "-structure=febmap")
            stru = SPEMFit::StFEBMAP;
        else {
            cerr << arg << endl;
            exitWithMsg("Invalid option detected.");
        }
        ix++;
    }

    if (input=="") {
        printUsage ();
        exit(1);
    }

    message ("Opening trace file... ", verbose);
    Trace* trace;
    if (multitype)
        trace = Trace::loadFromFile (input, 0, 1);
    else
        trace = Trace::loadFromFile (input);
    if (trace)
        message ("Trace file loaded.", verbose);
    else
        exitWithMsg("Loading of trace file failed.");

    auto tStart = chrono::system_clock::now();
    MixtureDensityMAP* map = NULL;

    message ("Starting SPEM-FIT.", verbose);
    if (verbose) {
        if (useBro) {
            cout << "Fitting with ";
            for (int i=0; i<branchOrders.size(); i++) {
                cout << branchOrders[i];
                if (i<branchOrders.size()-1)
                    cout << ",";
            }
            cout << " branches." << endl;
        }
        else
            cout << "Fitting with " << states << " states." << endl;
    }
    SPEMFit spemFit (trace);
    if (useBro)
        spemFit.initFitting (branchOrders, stru, initial);
    else
        spemFit.initFitting (states, strategy, stru, initial);
    map = spemFit.runFitting (ph, da, precision, maxiter, verbose);


    if (!map)
        exitWithMsg("Fitting failed.");

    auto tEnd = chrono::system_clock::now();
    auto ms = chrono::duration_cast<chrono::milliseconds>(tEnd-tStart).count();
    cout     << "=== Total time = " << ms << "ms ===" << endl;

    if (evaluate) {
        cout << endl;
        cout << "----------------------------------------" << endl;
        cout << "       Evaluation of the results:       " << endl;
        cout << "----------------------------------------" << endl;
        cout << "Log likelihood / trace size = " << endl;
        map->printSummary (cout);
        cout << "----------------------------------------" << endl;
        if (multitype)
            evaluateMultiType (trace, map, ph);
        else
            evaluateSingleType (trace, map, ph);
    }

    if (output!="") {
        ofstream file;
        file.open (output.c_str());
        if (file) {
            file << setprecision(18);
            if (ph)
                map->printAsPH (file, format);
            else
                map->printAsMAP (file, format);
            file.close();
        }
        else
            exitWithMsg("Can not open output file!");
    }
    delete trace;
    delete map;
}
