//  SPEM-Fit: a fast EM algorithm for MAP fitting
//  Copyright (C) 2013, Gabor Horvath
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "trace.h"
#include <fstream>
#include <cstdlib>
#include <algorithm>
#include <cmath>

using namespace std;

Trace* Trace::loadFromFile(const string &fname, int iaFieldNum, int typeFieldNum, const string &delimiters, bool incremental, int startLine, int endLine) {

    ifstream file;
    file.open(fname.c_str());

    if (!file)
        return NULL;

    Trace* res = new Trace;
    res->hasZero = false;
    res->numOfTypes = 1;

    string line;
    int ix = 0;
    // skip lines not required
    while (ix < startLine && getline (file, line))
        ix++;

    double olddata;
    bool first = true;
    while (getline (file, line) && (endLine==-1 || ix <= endLine)) {

        // extract the iatime field needed
        string::size_type lastPos = line.find_first_not_of(delimiters, 0);
        string::size_type pos = line.find_first_of(delimiters, lastPos);
        for (int fnum=0; fnum < iaFieldNum && (pos != string::npos || lastPos != string::npos); fnum++) {
            lastPos = line.find_first_not_of(delimiters, pos);
            pos = line.find_first_of(delimiters, lastPos);
        }
        double data = atof (line.substr(lastPos, pos - lastPos).c_str());
        if (data==0)
            res->hasZero = true;
        if (incremental) {
            if (!first)
                res->iatime.push_back(data - olddata);
            else
                first = false;
            olddata = data;
        }
        else
            res->iatime.push_back(data);

        // extract the type field
        if (typeFieldNum>=0) {
            string::size_type lastPos = line.find_first_not_of(delimiters, 0);
            string::size_type pos = line.find_first_of(delimiters, lastPos);
            for (int fnum=0; fnum < typeFieldNum && (pos != string::npos || lastPos != string::npos); fnum++) {
                lastPos = line.find_first_not_of(delimiters, pos);
                pos = line.find_first_of(delimiters, lastPos);
            }
            int t = atoi (line.substr(lastPos, pos - lastPos).c_str());
            res->numOfTypes = max(res->numOfTypes, t);
            res->type.push_back (t - 1);
        }

        ix++;
    }
    return res;
}

vector<double> Trace::getMoments(int upToOrder) const {

    vector<double> moms (upToOrder+1, 0.0);
    for (int i=0; i<iatime.size(); i++) {
        double t = 1.0;
        for (int m=0; m<=upToOrder; m++) {
            moms[m] += t;
            t *= iatime[i];
        }
    }
    for (int m=0; m<=upToOrder; m++)
        moms[m] /= iatime.size();
    return moms;
}

std::vector<double> Trace::getBackwardMoments(int upToOrder, int k) const {

    vector<double> moms (upToOrder+1, 0.0);
    if (type.empty())
        return moms;

    for (int i=0; i<iatime.size(); i++) {
        if (type[i]==k) {
            double t = 1.0;
            for (int m=0; m<=upToOrder; m++) {
                moms[m] += t;
                t *= iatime[i];
            }
        }
    }
    for (int m=upToOrder; m>=0; m--)
        moms[m] /= moms[0];
    return moms;
}

std::vector<double> Trace::getForwardMoments(int upToOrder, int k) const {

    vector<double> moms (upToOrder+1, 0.0);
    if (type.empty())
        return moms;

    for (int i=1; i<iatime.size(); i++) {
        if (type[i-1]==k) {
            double t = 1.0;
            for (int m=0; m<=upToOrder; m++) {
                moms[m] += t;
                t *= iatime[i];
            }
        }
    }
    for (int m=upToOrder; m>=0; m--)
        moms[m] /= moms[0];
    return moms;
}

vector<vector<double> > Trace::getJointMoments(int upToOrder) const {

    vector<vector<double> > moms (upToOrder+1);
    for (int i=0; i<=upToOrder; i++) {
        moms[i].resize(upToOrder+1);
        fill(moms[i].begin(), moms[i].end(), 0.0);
    }
    double ipow, jpow;
    for (int k=0; k<iatime.size()-1; k++) {
        ipow = 1.0;
        for (int i=0; i<=upToOrder; i++) {
            jpow = 1.0;
            for (int j=0; j<=upToOrder; j++) {
                moms[i][j] += ipow*jpow;
                jpow *= iatime[k+1];
            }
            ipow *= iatime[k];
        }
    }
    for (int i=0; i<=upToOrder; i++)
        for (int j=0; j<=upToOrder; j++)
            moms[i][j] /= (iatime.size()-1);
    return moms;
}

std::vector<std::vector<double> > Trace::getJointMomentsOfType(int upToOrder, int p) const {

    vector<vector<double> > moms (upToOrder+1);
    for (int i=0; i<=upToOrder; i++) {
        moms[i].resize(upToOrder+1);
        fill(moms[i].begin(), moms[i].end(), 0.0);
    }
    if (type.empty())
        return moms;

    double ipow, jpow;
    for (int k=0; k<iatime.size()-1; k++) {
        if (type[k]==p) {
            ipow = 1.0;
            for (int i=0; i<=upToOrder; i++) {
                jpow = 1.0;
                for (int j=0; j<=upToOrder; j++) {
                    moms[i][j] += ipow*jpow;
                    jpow *= iatime[k+1];
                }
                ipow *= iatime[k];
            }
        }
    }
    for (int i=0; i<=upToOrder; i++)
        for (int j=0; j<=upToOrder; j++)
            moms[i][j] /= (iatime.size()-1);
    return moms;
}

vector<double> Trace::getAcf(int maxLag) const {

    vector<double> res (maxLag);
    vector<double> moms = getMoments(3);
    double mean2 = moms[1] * moms[1];
    double var = moms[2] - mean2;

    for (int i=0; i<maxLag; i++) {
        for (int k=0; k<iatime.size()-i; k++)
            res[i] += iatime[k]*iatime[k+i] / (iatime.size()-i);
         res[i] = (res[i] - mean2) / var;
    }
    return res;
}

std::vector<double> Trace::getCrossAcf(int maxLag, int p1, int p2) const {

    vector<double> res (maxLag, 0.0);
    vector<double> moms1 = getBackwardMoments (2, p1);
    vector<double> moms2 = getBackwardMoments (2, p2);

    double std1 = sqrt (moms1[2] - moms1[1]*moms1[1]);
    double std2 = sqrt (moms2[2] - moms2[1]*moms2[1]);

    for (int i=0; i<maxLag; i++) {
        int num = 0;
        for (int k=0; k<iatime.size()-i; k++)
            if (type[k]==p1 && type[k+i]==p2) {
                res[i] += (iatime[k]-moms1[1])*(iatime[k+i]-moms2[1]);
                num++;
            }
        res[i] = res[i] / num / std1 / std2;
    }
    return res;
}

ostream& operator<< (ostream& out, const Trace& trace) {

    if (trace.isSingleType())
        for (int i=0; i<trace.size(); i++)
            out << trace.getIATime(i) << endl;
    else
        for (int i=0; i<trace.size(); i++)
            out << trace.getIATime(i) << '\t' << trace.getType(i) << endl;
    return out;
}

std::vector<char> Trace::kmeans(int numOfClusters) {

    if (clusters.find(numOfClusters)!=clusters.end())
        return clusters[numOfClusters];

    const vector<double>& entries = iatime;

    // sort data such that the original indices are preserved
    vector<pair<double,int> > valIdxPairs (entries.size());
    for (int i=0; i<entries.size(); i++)
        valIdxPairs[i] = make_pair(entries[i], i);
    sort (valIdxPairs.begin(), valIdxPairs.end());

    // set initial cluster means
    vector<double> currMeans (numOfClusters), oldMeans (numOfClusters);
    double min = valIdxPairs.front().first, max = valIdxPairs.back().first;
    for (int i=0; i<numOfClusters; i++)
        currMeans[i] = min + (max - min) * (2*i+1) / 2 / numOfClusters;

    vector<int> endBorders (numOfClusters);
    vector<char> clusterNum (entries.size());
    std::fill (clusterNum.begin(),clusterNum.end(), 0);

    bool ready = false;
    while (!ready) {
        // save current cluster means
        oldMeans = currMeans;

        // find end of clusters
        for (int k=0; k<numOfClusters-1; k++) {
            double limit = (currMeans[k+1] + currMeans[k]) / 2;
            // find border by interval bisectioning
            int lower = k==0 ? 0 : (endBorders[k-1] + 1);
            int upper = entries.size() - 1;
            while (lower<upper-1) {
                int test = (upper + lower) / 2;
                if (valIdxPairs[test].first <= limit)
                    lower = test;
                else
                    upper = test - 1;
            }
            if (valIdxPairs[upper].first <= limit)
                endBorders[k] = upper;
            else
                endBorders[k] = lower;
        }
        endBorders[numOfClusters-1] = entries.size() - 1;

        // calculate new cluster means
        int lower = 0;
        for (int k=0; k<numOfClusters; k++) {
            currMeans[k] = 0.0;
            int num = 0;
            for (int i=lower; i<=endBorders[k]; i++) {
                currMeans[k] += valIdxPairs[i].first;
                clusterNum[valIdxPairs[i].second] = k;
                num++;
            }
            currMeans[k] /= num;
            lower = endBorders[k]+1;
        }

        // check if we can stop
        ready = true;
        for (int i=0; i<currMeans.size(); i++)
            if (fabs(currMeans[i]-oldMeans[i]) > 1e-14) {
                ready = false;
                break;
            }
    }
    clusters[numOfClusters] = clusterNum;
    return clusterNum;
}



